/* This is a generated file, edit the .stub.php file instead.
 * Stub hash: 598f048ec627ac0df4af6941b0e598772908c7cb */

static zend_class_entry *register_class_MongoDB_BSON_MinKeyInterface(void)
{
	zend_class_entry ce, *class_entry;

	INIT_NS_CLASS_ENTRY(ce, "MongoDB\\BSON", "MinKeyInterface", NULL);
	class_entry = zend_register_internal_interface(&ce);

	return class_entry;
}
