/* This is a generated file, edit the .stub.php file instead.
 * Stub hash: c3b36fa7e01da6c3e30280afc23e71b81ef39c36 */

ZEND_BEGIN_ARG_INFO_EX(arginfo_class_MongoDB_Driver_BulkWriteCommandResult___construct, 0, 0, 0)
ZEND_END_ARG_INFO()

ZEND_BEGIN_ARG_WITH_RETURN_TYPE_INFO_EX(arginfo_class_MongoDB_Driver_BulkWriteCommandResult_getInsertedCount, 0, 0, IS_LONG, 0)
ZEND_END_ARG_INFO()

#define arginfo_class_MongoDB_Driver_BulkWriteCommandResult_getMatchedCount arginfo_class_MongoDB_Driver_BulkWriteCommandResult_getInsertedCount

#define arginfo_class_MongoDB_Driver_BulkWriteCommandResult_getModifiedCount arginfo_class_MongoDB_Driver_BulkWriteCommandResult_getInsertedCount

#define arginfo_class_MongoDB_Driver_BulkWriteCommandResult_getUpsertedCount arginfo_class_MongoDB_Driver_BulkWriteCommandResult_getInsertedCount

#define arginfo_class_MongoDB_Driver_BulkWriteCommandResult_getDeletedCount arginfo_class_MongoDB_Driver_BulkWriteCommandResult_getInsertedCount

ZEND_BEGIN_ARG_WITH_RETURN_OBJ_INFO_EX(arginfo_class_MongoDB_Driver_BulkWriteCommandResult_getInsertResults, 0, 0, MongoDB\\BSON\\Document, 1)
ZEND_END_ARG_INFO()

#define arginfo_class_MongoDB_Driver_BulkWriteCommandResult_getUpdateResults arginfo_class_MongoDB_Driver_BulkWriteCommandResult_getInsertResults

#define arginfo_class_MongoDB_Driver_BulkWriteCommandResult_getDeleteResults arginfo_class_MongoDB_Driver_BulkWriteCommandResult_getInsertResults

ZEND_BEGIN_ARG_WITH_RETURN_TYPE_INFO_EX(arginfo_class_MongoDB_Driver_BulkWriteCommandResult_isAcknowledged, 0, 0, _IS_BOOL, 0)
ZEND_END_ARG_INFO()

static ZEND_METHOD(MongoDB_Driver_BulkWriteCommandResult, __construct);
static ZEND_METHOD(MongoDB_Driver_BulkWriteCommandResult, getInsertedCount);
static ZEND_METHOD(MongoDB_Driver_BulkWriteCommandResult, getMatchedCount);
static ZEND_METHOD(MongoDB_Driver_BulkWriteCommandResult, getModifiedCount);
static ZEND_METHOD(MongoDB_Driver_BulkWriteCommandResult, getUpsertedCount);
static ZEND_METHOD(MongoDB_Driver_BulkWriteCommandResult, getDeletedCount);
static ZEND_METHOD(MongoDB_Driver_BulkWriteCommandResult, getInsertResults);
static ZEND_METHOD(MongoDB_Driver_BulkWriteCommandResult, getUpdateResults);
static ZEND_METHOD(MongoDB_Driver_BulkWriteCommandResult, getDeleteResults);
static ZEND_METHOD(MongoDB_Driver_BulkWriteCommandResult, isAcknowledged);

static const zend_function_entry class_MongoDB_Driver_BulkWriteCommandResult_methods[] = {
	ZEND_ME(MongoDB_Driver_BulkWriteCommandResult, __construct, arginfo_class_MongoDB_Driver_BulkWriteCommandResult___construct, ZEND_ACC_PRIVATE|ZEND_ACC_FINAL)
	ZEND_ME(MongoDB_Driver_BulkWriteCommandResult, getInsertedCount, arginfo_class_MongoDB_Driver_BulkWriteCommandResult_getInsertedCount, ZEND_ACC_PUBLIC|ZEND_ACC_FINAL)
	ZEND_ME(MongoDB_Driver_BulkWriteCommandResult, getMatchedCount, arginfo_class_MongoDB_Driver_BulkWriteCommandResult_getMatchedCount, ZEND_ACC_PUBLIC|ZEND_ACC_FINAL)
	ZEND_ME(MongoDB_Driver_BulkWriteCommandResult, getModifiedCount, arginfo_class_MongoDB_Driver_BulkWriteCommandResult_getModifiedCount, ZEND_ACC_PUBLIC|ZEND_ACC_FINAL)
	ZEND_ME(MongoDB_Driver_BulkWriteCommandResult, getUpsertedCount, arginfo_class_MongoDB_Driver_BulkWriteCommandResult_getUpsertedCount, ZEND_ACC_PUBLIC|ZEND_ACC_FINAL)
	ZEND_ME(MongoDB_Driver_BulkWriteCommandResult, getDeletedCount, arginfo_class_MongoDB_Driver_BulkWriteCommandResult_getDeletedCount, ZEND_ACC_PUBLIC|ZEND_ACC_FINAL)
	ZEND_ME(MongoDB_Driver_BulkWriteCommandResult, getInsertResults, arginfo_class_MongoDB_Driver_BulkWriteCommandResult_getInsertResults, ZEND_ACC_PUBLIC|ZEND_ACC_FINAL)
	ZEND_ME(MongoDB_Driver_BulkWriteCommandResult, getUpdateResults, arginfo_class_MongoDB_Driver_BulkWriteCommandResult_getUpdateResults, ZEND_ACC_PUBLIC|ZEND_ACC_FINAL)
	ZEND_ME(MongoDB_Driver_BulkWriteCommandResult, getDeleteResults, arginfo_class_MongoDB_Driver_BulkWriteCommandResult_getDeleteResults, ZEND_ACC_PUBLIC|ZEND_ACC_FINAL)
	ZEND_ME(MongoDB_Driver_BulkWriteCommandResult, isAcknowledged, arginfo_class_MongoDB_Driver_BulkWriteCommandResult_isAcknowledged, ZEND_ACC_PUBLIC|ZEND_ACC_FINAL)
	ZEND_FE_END
};

static zend_class_entry *register_class_MongoDB_Driver_BulkWriteCommandResult(void)
{
	zend_class_entry ce, *class_entry;

	INIT_NS_CLASS_ENTRY(ce, "MongoDB\\Driver", "BulkWriteCommandResult", class_MongoDB_Driver_BulkWriteCommandResult_methods);
#if (PHP_VERSION_ID >= 80400)
	class_entry = zend_register_internal_class_with_flags(&ce, NULL, ZEND_ACC_FINAL|ZEND_ACC_NOT_SERIALIZABLE);
#else
	class_entry = zend_register_internal_class_ex(&ce, NULL);
	class_entry->ce_flags |= ZEND_ACC_FINAL|ZEND_ACC_NOT_SERIALIZABLE;
#endif

	return class_entry;
}
