/* This is a generated file, edit the .stub.php file instead.
 * Stub hash: a86f5de0f86037e928a37f1b6c38593ce005b202 */

static zend_class_entry *register_class_MongoDB_Driver_Monitoring_Subscriber(void)
{
	zend_class_entry ce, *class_entry;

	INIT_NS_CLASS_ENTRY(ce, "MongoDB\\Driver\\Monitoring", "Subscriber", NULL);
	class_entry = zend_register_internal_interface(&ce);

	return class_entry;
}
