/* This is a generated file, edit the .stub.php file instead.
 * Stub hash: 7c297c1f6a9454b4e7a509b278f2776e31d8dbfa */

static zend_class_entry *register_class_MongoDB_BSON_Type(void)
{
	zend_class_entry ce, *class_entry;

	INIT_NS_CLASS_ENTRY(ce, "MongoDB\\BSON", "Type", NULL);
	class_entry = zend_register_internal_interface(&ce);

	return class_entry;
}
